from metafind.client import Client
from metafind.models import Document

from pathlib import Path
from exiftool.exceptions import ExifToolOutputEmptyError

import exiftool


_TAGS_TO_SKIP = {"SourceFile", "ExifTool:ExifToolVersion", "ExifTool:Warning"}


class ExifToolClient(Client):
    """
    ExifToolClient is a wrapper built around the `exiftool.ExifToolHelper` class.
    """

    def __init__(self, path: str):
        self.client = exiftool.ExifToolHelper()
        self.path = path

    def get_metadata(self) -> list[Document]:
        result = self.client.get_metadata(self.path)
        return [_make_new_doc(data) for data in result]

    def scrub_metadata(self) -> (list[str], list[str]):
        all_tags = self.get_unique_tags()
        try:
            self.client.execute_json("-all= ", self.path)
        except ExifToolOutputEmptyError:
            immutable_tags = self.get_unique_tags()
            mutable_tags = all_tags - immutable_tags
            return sorted(mutable_tags), sorted(immutable_tags)

    def get_unique_tags(self) -> set[str]:
        result = self.client.get_metadata(self.path)
        unique_tags = {
            tag.split(":", 1)[1]
            for data in result
            for tag in data
            if tag not in _TAGS_TO_SKIP
        }
        return unique_tags


def _make_new_doc(data: dict) -> Document:
    name = Path(data["SourceFile"]).name
    filetype = data["File:FileType"]
    return Document(name, data, filetype, _TAGS_TO_SKIP)



class MetafindError(Exception):
    """
    Base class for all metafind exceptions.
    """

    def __init__(self, message, error_code):
        self.message = message
        self.error_code = error_code

    def __str__(self) -> str:
        return f"{self.message}, {self.error_code}"


class UnsupportedBackend(MetafindError):
    """
    Exception that's raised when an unsupported backend is specified.
    """

    ERROR_CODE = 1

    def __init__(self, backend: str):
        super().__init__(
            f"{backend} is not currently supported as a backend for metafind",
            self.ERROR_CODE,
        )


class BackendNotFound(MetafindError):
    """
    Exception that's raised when the specified backend could not be located on the host.
    """

    ERROR_CODE = 2

    def __init__(self, backend: str):
        super().__init__(f"{backend} could not be found", self.ERROR_CODE)



from metafind.client import Client
from metafind.exifclient import ExifToolClient
from metafind.errors import UnsupportedBackend, BackendNotFound


from shutil import which


def get_client(backend: str, path: str) -> Client:
    """
    get_client returns a client for a metadata backend that has a standardized API for performing metadata analysis tasks.
    """
    if not _backend_exists(backend):
        raise BackendNotFound(backend)

    match backend:
        case "exiftool":
            return ExifToolClient(path)
        case _:
            raise UnsupportedBackend(backend)


def _backend_exists(backend: str):
    """
    backend_exists checks if the backend is on the host.

    Current implementation is naive, and assumes the provided name of the backend is the name of the tool on the system path.
    Should be more robust in the future, however this is the lowest amount of code implementation :)
    """
    if which(backend):
        return True
    return False



from abc import ABC, abstractmethod


class Client(ABC):
    """
    Client is the base class that all metadata clients will inherent and implement.
    This is used to enable the support of additional backends, which will be injected at runtime into the caller classes.
    """

    @abstractmethod
    def get_metadata(self):
        pass

    @abstractmethod
    def scrub_metadata(self):
        pass

    @abstractmethod
    def get_unique_tags(self):
        pass



from metafind.client import Client
from metafind.clientfactory import get_client
from metafind.errors import MetafindError
from metafind.output import fetch_output

from rich.console import Console
from rich.logging import RichHandler


import sys
import click
import signal
import logging

_SIGINT_EXIT_CODE = 130

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)

log = logging.getLogger("rich")


def common_options(func):
    """
    Decorator to attach the same CLI options ('path', 'backend')
    to each command.
    """
    func = click.argument("path", required=True)(func)
    func = click.option(
        "--backend",
        help="Specify a metadata analyzer backend",
        default="exiftool",
        required=True,
    )(func)
    return func


def handle_command(func):
    """
    Decorator to create the backend client and CLI console, as well as handle any exceptions and signals.
    """

    def wrapper(path, backend, *args, **kwargs):
        signal.signal(signal.SIGINT, handle_sigint)
        try:
            client = get_client(backend, path)
            console = Console()
            return func(client, console, *args, **kwargs)
        except MetafindError as e:
            sys.exit(e.error_code)

    # Preserve function metadata (so click doesn't get confused)
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


def handle_sigint(signum, frame):
    """
    Handler to gracefully exit when sigint is encountered.
    """
    sys.exit(_SIGINT_EXIT_CODE)


@click.group()
def cli():
    """
    Metadata analysis tool created for CSEC-473 - Penetration Testing.

    Authored by sapph2c
    """
    pass


@cli.command(short_help="get file(s) metadata")
@common_options
@handle_command
def fetch(client: Client, console: Console) -> None:
    """
    Retrieves a list of all metadata found within the file(s) at the specified path.
    """
    result = client.get_metadata()
    console.print(fetch_output(result))


@cli.command(short_help="scrub file(s) metadata")
@common_options
@handle_command
def scrub(client: Client, console: Console):
    """
    Attempts to scrub all metadata found within the file(s) at the specified path.
    """
    mutable_tags, immutable_tags = client.scrub_metadata()
    console.print("-------- mutable tags --------")
    console.print(sorted(mutable_tags))
    console.print("-------- immutable tags --------")
    console.print(sorted(immutable_tags))


@cli.command(short_help="get unique tags from file(s)")
@common_options
@handle_command
def unique(client: Client, console: Console) -> None:
    """
    Gets an alphabetized list of all unique metadata tags found within the file(s) at the specified path.
    """
    result = client.get_unique_tags()
    console.print(sorted(result))



from metafind.models import Document
from rich.table import Table
from rich.console import Group


def fetch_output(documents: list[Document]):
    tables = create_document_tables(documents)
    return Group(*tables)


def create_document_tables(documents: list[Document]) -> list[Table]:
    tables: list[Table] = []
    for doc in documents:
        table = Table(title=doc.name)
        table.add_column("Tag", style="cyan", no_wrap=True)
        table.add_column("Value", style="magenta")

        # Add rows for each key-value pair (except SourceFile itself if desired)
        for data in doc.metadata:
            table.add_row(data.tag, str(data.value))
        tables.append(table)
    return tables



from dataclasses import dataclass, field


@dataclass(order=True)
class Metadata:
    """
    Metadata is a class used to encapsulate discovered metadata.

    :param parent: Identifies the parent file or resource from which the metadata was discovered.
    :type parent: str
    :param tag: The metadata tag (used for sorting).
    :type tag: str
    :param value: The value associated with the given tag.
    :type value: str
    """

    parent: str = field(compare=False)
    tag: str = field(compare=True)
    value: str = field(compare=False)


class Document:
    """
    Document is a class used to encapsulate file metadata.
    """

    def __init__(self, name: str, data: dict, filetype: str, tags_to_skip: list):
        self.name = name
        self.filetype = filetype
        self.metadata: list[Metadata] = self._add_metadata(data, tags_to_skip)

    def _add_metadata(self, data: dict, tags_to_skip: list) -> list[Metadata]:
        """
        add_metadata adds metadata to the Document.
        """
        metadata: list[Metadata] = []
        for tag, value in data.items():
            if tag not in tags_to_skip:
                clean_tag = tag.split(":")[1]
                metadata.append(Metadata(parent=self.name, tag=clean_tag, value=value))
        return metadata

    def __repr__(self) -> str:
        return f"{self.name}, {self.filetype}, {self.metadata}"
